package edu.uiuc.dsi.mutex;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.PriorityQueue;

import edu.uiuc.dsi.Packet;
import edu.uiuc.dsi.TCPSocket;
import edu.uiuc.dsi.clocks.LamportClock;

public class Lamport extends edu.uiuc.dsi.mutex.Base
{
	// Lamport's MutEx Algorithm
	//
	// Requesting site P_i broadcasts Request(TS)
	// Requesting site adds Request(TS, i) to request queue
	//
	// Upon Request(TS), sends Reply(TS)
	// Places Request(TS, i) onto its request queue
	//
	// Header: 0x01: Request
	//         0x02: Reply
	private static final byte MESSAGE_REQUEST = (byte)0x01;
	private static final byte MESSAGE_REPLY = (byte)0x02;
	private static final byte MESSAGE_RELEASE = (byte)0x03;
	
	private volatile LamportClock objLamportClock = new LamportClock();
	private PriorityQueue<LamportQueueElement> objLamportQueue = new PriorityQueue<LamportQueueElement>();
	private ArrayList<TCPSocket> objTCPSocketsUnreplied = new ArrayList<TCPSocket>();

	public Lamport(String componentName)
	{
		super(componentName);
	}
	
	@Override
	public void processPacket(TCPSocket sender, Packet packet)
	{
		ByteBuffer objByteBuffer = ByteBuffer.wrap(packet.getData()), objResponseByteBuffer;
		byte packetType = objByteBuffer.get();
		short packetTimestamp = 0;

		// 1. Update the local clock
		objLamportClock.tick();
		
		// 2. Process the packet
		switch (packetType)
		{
			case MESSAGE_REQUEST:
				// 1. Update the local Lamport clock
				packetTimestamp = objByteBuffer.getShort();
				objLamportClock.tick(packetTimestamp);
				
				// 2. Add the REQUEST to the REQUEST queue
				objLamportQueue.add(new LamportQueueElement(packetTimestamp, sender));
				
				// 3. Send the REPLY message to REQUESTer
				objResponseByteBuffer = ByteBuffer.allocate(3);
				objResponseByteBuffer.put(MESSAGE_REPLY);
				objResponseByteBuffer.putShort(objLamportClock.getCurrentTime());
				write(sender, objResponseByteBuffer.array());
				
				break;
				
			case MESSAGE_REPLY:
				// 1. Update the local Lamport clock
				packetTimestamp = objByteBuffer.getShort();
				objLamportClock.tick(packetTimestamp);
				
				// 2. Remove the sender of the REPLY from the necessary REPLIES needed 
				objTCPSocketsUnreplied.remove(sender);
				
				break;
				
			case MESSAGE_RELEASE:
				// 1. Remove the top object from the queue
				objLamportQueue.remove();
				
				break;
		}

		if (!objLamportQueue.isEmpty() && objTCPSocketsUnreplied.size() == 0 && objLamportQueue.peek().getSiteTCPSocket() == getSelfTCPSocket() && getCurrentMutualExclusionState() == Base.REQUESTING_CRITICAL_SECTION)
		{
			try { beginMutEx(); }
			catch (MutExException e) { e.printStackTrace(); }
		}
	}

	@Override
	protected void doBeginMutExTasks()
	{
		objLamportClock.tick();
		
	}

	@Override
	protected void doReleaseMutExTasks()
	{
		objLamportClock.tick();
		
		ByteBuffer objByteBuffer = ByteBuffer.allocate(1);
		objByteBuffer.put(MESSAGE_RELEASE);
		broadcast(objByteBuffer.array());
	}
	
	@Override
	protected void doRequestMutExTasks()
	{
		objLamportClock.tick();
		
		// Mark all sites as needing to be heard from... 
		objTCPSocketsUnreplied.add(getSelfTCPSocket());
		for (TCPSocket objTCPSocket : getActiveTCPSockets())
			objTCPSocketsUnreplied.add(objTCPSocket); 
		
		// Broadcast request to all connected sites...
		ByteBuffer objByteBuffer = ByteBuffer.allocate(3);
		objByteBuffer.put(MESSAGE_REQUEST);
		objByteBuffer.putShort(objLamportClock.getCurrentTime());
		broadcast(objByteBuffer.array());
	}
}
